<?php

class EveryTwig extends Twig_Environment {

    private static $twig = null;
    private static $instance;
    private static $loaders = [ ];

    protected $cache = false;
    private $cache_time = 60;

    // Add suffix so we don't have to delete it in deploy every time
    private $cache_key = 'everytwig_compiled_cache_'.GIT_COMMIT;

    public static function get_instance() {
        if ( null == self::$instance ) {
            self::$instance = new self;
        }

        return self::$instance;
    }

    /**
     * @param array $loaders - array of Twig_LoaderInterface
     * @param array $functions
     * @param array $filters
     * @param array $options - array of Twig options
     *
     * @return EveryTwig instance
     *
     * Initializes the EveryTwig-wrapper and Twig environment
     */
    public function init( array $loaders, array $functions = [], array $filters = [], array $options = ['debug' => false, 'cache' => false] ) {

        // Handle options
        $this->handle_options( $options );

        foreach ( $loaders as $loader ) {
            $this->add_initial_loader( $loader );
        }

        if( $this->cache ) {
            if ( false === get_transient( $this->cache_key ) ) {
                set_transient( $this->cache_key, array(), $this->cache_time );
            }
        }

        $chain = new Twig_Loader_Chain( self::$loaders );
        $twig  = new Twig_Environment( $chain, $options );

        self::$twig = $twig;
        self::$twig->addFunction( $this->default_php_function() );

        foreach ( $functions as $function ) {
            self::$twig->addFunction( $function );
        }

        foreach ( $filters as $filter ) {
            self::$twig->addFilter( $filter );
        }

        self::$twig->setLexer( new \Twig_Lexer( $this ) );
        self::$twig->setParser( new \Twig_Parser( $this ) );

        return self::$instance;
    }

    /**
     * @param $options - array
     *
     * Handles the specific cache options and trims them away from the
     * options array.
     */
    private function handle_options( &$options ) {
        if ( isset( $options['cache'] ) ) {
            $this->cache = $options['cache'];
            unset( $options['cache'] );
        }

        if ( isset( $options['cachetime'] ) && intval( $options['cachetime'] ) > 0 ) {
            $this->cache_time = intval( $options['cachetime'] );
            unset( $options['cachetime'] );
        }
    }

    /**
     * Add initial loader
     *
     * @param Twig_LoaderInterface $loader
     *
     * Helper function to add a Twig_LoaderInterface loader on init
     */
    private function add_initial_loader( Twig_LoaderInterface $loader ) {
        array_push( self::$loaders, $loader );
    }

    /**
     * @param Twig_LoaderInterface $loader
     *
     * @throws Exception
     * @internal param array $loaders Function to add loaders after a Twig environment has been initialized*
     * Function to add loaders after a Twig environment has been initialized
     */
    public function add_loader( Twig_LoaderInterface $loader ) {
        if( is_null(self::$twig) ) {
            throw new Exception('Function init needs to be run in order to add extra loaders');
        }

        array_push(self::$loaders, $loader);
        self::$twig->setLoader( new Twig_Loader_Chain(self::$loaders) );
    }

    /**
     * @return Twig_SimpleFunction
     *
     * Twig_SimpleFunction to be able to call default php functions from within twig templates
     */
    protected function default_php_function() {
        $phpFunction = new Twig_SimpleFunction( 'php_function', function ( $name, $args = '' ) {
            if ( function_exists( $name ) ) {
                return call_user_func_array( $name, $args );
            }

            if ( strpos( $name, '::' ) ) {
                $static_method = explode( '::', $name );
                if ( method_exists( $static_method[0], $static_method[1] ) ) {
                    return call_user_func_array( $name, $args );
                }
            }

        } );

        return $phpFunction;
    }

    /**
     * @param string $template
     * @param array  $context
     *
     * @return string
     *
     * Render function of EveryTwig, loads a specific template and works with
     * cached compiled templates.
     */
    public function render( $template, array $context = array() ) {
        echo $this->loadTemplate( $template )->render( $context );
    }

    /**
     * Return the view as rendered HTML by passed template
     *
     * @param string $template
     * @param array  $context
     *
     * @return mixed
     */
    public function view( $template, array $context = array() ) {
        return $this->loadTemplate( $template )->render( $context );
    }

    /**
     * @param $name
     *
     * @return mixed
     *
     * Loads a specific template and compiles it. If cache is on, save the compiled class
     * in cac3he to be evaluated later on.
     */
    public function loadTemplate( $name, $index = null ) {
        $cls = self::$twig->getTemplateClass( $name, $index );

        if ( isset( self::$twig->loadedTemplates[ $cls ] ) ) {
            return self::$twig->loadedTemplates[ $cls ];
        }

        if ( ! class_exists( $cls, false ) ) {
            if ( false === ( $cache = $this->get_cached_template( $cls ) ) ) {
                $source = self::$twig->compileSource( self::$twig->getLoader()->getSource( $name ), $name );
                eval( '?>' . $source );

                if ( $this->cache ) {
                    $this->cache_template( $cls, serialize( $source ) );
                }
            } else {
                eval( '?>' . unserialize( $cache ) );
            }
        }

        if ( ! self::$twig->runtimeInitialized ) {
            self::$twig->initRuntime();
        }

        return self::$twig->loadedTemplates[ $cls ] = new $cls( $this );
    }

    /**
     * @param $name
     *
     * @return bool|mixed
     *
     * Helper function to fetch cached template
     */
    private function get_cached_template( $name ) {
        if ( ! $this->cache ) {
            return false;
        }

        $cache = get_transient( $this->cache_key );

        return $cache !== false && isset( $cache[ $name ] ) ? $cache[ $name ] : false;
    }

    /**
     * @param $name
     * @param $data
     *
     * Helper function to cache template
     */
    private function cache_template( $name, $data ) {
        $cache          = get_transient( $this->cache_key );
        $cache[ $name ] = $data;
        set_transient( $this->cache_key, $cache, $this->cache_time );
    }

    public function getTokenParsers() {
        return self::$twig->getTokenParsers();
    }

    public function getNodeVisitors() {
        return self::$twig->getNodeVisitors();
    }

    public function getUnaryOperators() {
        return self::$twig->getUnaryOperators();
    }

    public function getBinaryOperators() {
        return self::$twig->getBinaryOperators();
    }

    public function computeAlternatives( $name, $items ) {
        return self::$twig->computeAlternatives( $name, $items );
    }

    public function getTags() {
        return self::$twig->getTags();
    }

    public function getFunction( $name ) {
        return self::$twig->getFunction( $name );
    }

    public function getFunctions() {
        return self::$twig->getFunctions();
    }

    public function getBaseTemplateClass() {
        return self::$twig->getBaseTemplateClass();
    }

    public function getExtension( $name ) {
        return self::$twig->getExtension( $name );
    }

    public function getExtensions() {
        return self::$twig->getExtensions();
    }

    public function getFilter( $name ) {
        return self::$twig->getFilter( $name );
    }

    public function getFilters() {
        return self::$twig->getFilters();
    }


}
